unit uStrings;

interface

uses
  Windows, SysUtils;

procedure CreateMemoryContext(Len: Word; out dc: HDC; out bmp: HBITMAP);
procedure DestroyMemoryContext(dc: HDC; bmp: HBITMAP);
function CompareMemoryContext(dc1, dc2: HDC; Len: Word): Boolean;

procedure StoreString(dc: HDC; const Value: WideString);
procedure StoreAndCryptString(dc: HDC; const Value: WideString; const KeyName: PWideChar);
procedure RetrieveAndDeryptString(dc: HDC; out Value: WideString; Len: Word; const KeyName: PWideChar);

function CryptString(const Value: WideString; const KeyName: PWideChar): WideString;

function CollapseString(const Value: WideString; Len: Word): WideString;

function StringToHex(const Value: WideString): AnsiString;
function HexToString(const Value: AnsiString): WideString;

implementation

{******************************************************************************} 

function CharBit(Char: WideChar; Bit: Word): ByteBool; register;
asm
  bt Char,Bit
  setc al
end;

function CryptString(const Value: WideString; const KeyName: PWideChar): WideString;
var
  keydc: HDC;
  keybmp: HBITMAP;
  dc: HDC;
  bmp: HBITMAP;
  w, i, b: Integer;
  color: COLORREF;
  char: WideChar;
begin
  if (Length(Value) = 0) or (Length(Value) > 255) then
    Exit;

  Result := '';

  keydc := CreateCompatibleDC(0);
  if (keydc = 0) then
    Exit;
  try
    keybmp := LoadImageW(hInstance, KeyName, IMAGE_BITMAP, 0, 0, LR_MONOCHROME);
    if (keybmp = 0) then
      Exit;
    try
      if (SelectObject(keydc, keybmp) = 0) then
        Exit;

      dc := CreateCompatibleDC(0);
      if (dc = 0) then
        Exit;
      try
        w := Length(Value) shl 4;                    
        bmp := CreateBitmap(w, 1, 1, 1, nil);
        if (bmp = 0) then
          Exit;
        try
          if (SelectObject(dc, bmp) = 0) then
            Exit;

          for i := 0 to Length(Value) - 1 do
          begin
            for b := 0 to 15 do
            begin
              color := $ffffff;
              if CharBit(Value[i + 1], b) then color := 0;
              SetPixel(dc, (i shl 4) + b, 0, color);
            end;
          end;
                      
          BitBlt(dc, 0, 0, w, 1, keydc, 0, 0, SRCINVERT);

          for i := 0 to Length(Value) - 1 do
          begin
            char := WideChar(Word(0));
            for b := 0 to 15 do
            begin
              color := GetPixel(dc, (i shl 4) + b, 0);
              if (color = 0) then
                char := WideChar(Word(char) or (1 shl b));
            end;
            Result := Result + char;
          end;
        finally
          DeleteObject(bmp);
        end;
      finally
        DeleteDC(dc);
      end;

    finally
      DeleteObject(keybmp);
    end;
  finally
    DeleteDC(keydc);
  end;
end;

procedure CreateMemoryContext(Len: Word; out dc: HDC; out bmp: HBITMAP);
begin
  dc := CreateCompatibleDC(0);
  bmp := CreateBitmap(Len shl 4, 1, 1, 1, nil);
  SelectObject(dc, bmp);
end;

procedure DestroyMemoryContext(dc: HDC; bmp: HBITMAP);
begin
  DeleteObject(bmp);
  DeleteDC(dc);
end;

function CompareMemoryContext(dc1, dc2: HDC; Len: Word): Boolean;
var
  i: Integer;
begin
  Result := FALSE;
  for i := 0 to Len shl 4 do
    if GetPixel(dc1, i, 0) <> GetPixel(dc2, i, 0) then Exit; 
  Result := TRUE;
end;

procedure StoreString(dc: HDC; const Value: WideString);
var
  i, b: Integer;
  color: COLORREF;
begin
  if (Length(Value) = 0) or (Length(Value) > 255) then
    Exit;

  for i := 0 to Length(Value) - 1 do
  begin
    for b := 0 to 15 do
    begin
      color := $ffffff;
      if CharBit(Value[i + 1], b) then color := 0;
      SetPixel(dc, (i shl 4) + b, 0, color);
    end;
  end;
end;

procedure StoreAndCryptString(dc: HDC; const Value: WideString; const KeyName: PWideChar);
var
  keydc: HDC;
  keybmp: HBITMAP;
  i, b: Integer;
  color: COLORREF;
begin
  if (Length(Value) = 0) or (Length(Value) > 255) then
    Exit;

  keydc := CreateCompatibleDC(0);
  if (keydc = 0) then
    Exit;
  try
    keybmp := LoadImageW(hInstance, KeyName, IMAGE_BITMAP, 0, 0, LR_MONOCHROME);
    if (keybmp = 0) then
      Exit;
    try
      if (SelectObject(keydc, keybmp) = 0) then
        Exit;

        for i := 0 to Length(Value) - 1 do
        begin
          for b := 0 to 15 do
          begin
            color := $ffffff;
            if CharBit(Value[i + 1], b) then color := 0;
            SetPixel(dc, (i shl 4) + b, 0, color);
          end;
        end;
                      
        BitBlt(dc, 0, 0, Length(Value) shl 4, 1, keydc, 0, 0, SRCINVERT);
    finally
      DeleteObject(keybmp);
    end;                          
  finally
    DeleteDC(keydc);
  end;
end;

procedure RetrieveAndDeryptString(dc: HDC; out Value: WideString; Len: Word; const KeyName: PWideChar);
var
  keydc: HDC;
  keybmp: HBITMAP;
  i, b: Integer;
  color: COLORREF;
  char: WideChar;
begin
  keydc := CreateCompatibleDC(0);
  if (keydc = 0) then
    Exit;
  try
    keybmp := LoadImageW(hInstance, KeyName, IMAGE_BITMAP, 0, 0, LR_MONOCHROME);
    if (keybmp = 0) then
      Exit;
    try
      if (SelectObject(keydc, keybmp) = 0) then
        Exit;         

        BitBlt(dc, 0, 0, Len shl 4, 1, keydc, 0, 0, SRCINVERT);
        
        Value := '';
        for i := 0 to Len - 1 do
        begin
          char := WideChar(Word(0));
          for b := 0 to 15 do
          begin
            color := GetPixel(dc, (i shl 4) + b, 0);
            if (color = 0) then
              char := WideChar(Word(char) or (1 shl b));
          end;
          Value := Value + char;
        end;
    finally
      DeleteObject(keybmp);
    end;
  finally
    DeleteDC(keydc);
  end;
end;

function CollapseString(const Value: WideString; Len: Word): WideString;
var
  i, n: Word;
begin
  Result := '';
  if (Length(Value) < Len) then Exit;
  if (Length(Value) = Len) then
  begin
    Result := Value;
    Exit;
  end;
  Result := Value;
  SetLength(Result, Len);

  n := 0;
  for i := 1 to Length(Value) do
  begin
    if (n >= Len) then
    begin
      n := 1;
    end else
      Inc(n);

    Result[n] := WideChar(Word(Result[n]) + Word(Value[i]));
  end;
end;

function StringToHex(const Value: WideString): AnsiString;
const
  convertArray: array[0..15] of Char = '0123456789ABCDEF';
var
  i: Word;
begin
  Result := '';
  for i := 1 to Length(Value) do
  begin
    Result := Result +
      convertArray[(Word(Value[i]) and $ff) shr 4] +
      convertArray[(Word(Value[i]) and $ff) and 15] +
      convertArray[(Word(Value[i]) shr 8) shr 4] +
      convertArray[(Word(Value[i]) shr 8) and 15];
  end;
end;

function HexToString(const Value: AnsiString): WideString;
var
  i: Word;
  w: Word;
begin
  Result := '';
  try
    for i := 0 to (Length(Value) shr 2)-1 do
    begin
      w := Word(StrToInt('$' +
        Value[(i shl 2)+3] + Value[(i shl 2)+4] +
        Value[(i shl 2)+1] + Value[(i shl 2)+2]));
      Result := Result + WideChar(w);
    end;
  except
    Result := '';
  end;
end;

end.
